/*	ap_mitm.c  --  Add Man-in-the-Middle mode to apache proxy module.
**
**	This is intended to let unmodified clients (e.g. Netscape) use
**	kerberized ssl via a local proxy.  The client talks ssl to the
**	local proxy using a fast, less secure cipher while the proxy
**	talks to the kssl server using a KRB5-* cipher.
**
**	These routines will be called in src/modules/proxy/proxy_connect.c
**	ap_proxy_connect_handler().
**
**	This code is based heavily on stunnel, (c) 1998-2000 Michal Trojnara
**	and consequently falls under GNU General Public License, version 2,
**	or later.
**
**	20000808 VRS 	Started.
*/

#include <sys/types.h> 
#include <stdio.h>
#include <errno.h>    
#include <stdlib.h>
#include <string.h>  
#include <ctype.h>
#include <sys/stat.h>

#include "ap_mitm.h"
#include "../ssl/mod_ssl.h"


void ssl3_send_alert(SSL *s,int level, int desc);

/* Correct callback definitions overriding ssl.h */
#ifndef NO_RSA
#ifdef SSL_CTX_set_tmp_rsa_callback
	#undef SSL_CTX_set_tmp_rsa_callback
#endif
#define SSL_CTX_set_tmp_rsa_callback(ctx,cb) \
	SSL_CTX_ctrl(ctx,SSL_CTRL_SET_TMP_RSA_CB,0,(char *)cb)
#endif /* NO_RSA */

#ifndef NO_RSA
	RSA 	*rsa_tmp;           /* temporary RSA key */
	static RSA *tmp_rsa_cb(SSL *s, int export, int keylength)
	{	/* temporary RSA key callback */
		/* log(LOG_DEBUG, "Returned temporary RSA callback");  */
		return rsa_tmp;
	}
#endif /* NO_RSA */


/*	Create and return an SSL context.
**	Calling with server==0 creates an ssl client context,
**	otherwise creates an ssl server context.
*/
SSL_CTX	*context_init(server_rec *s, int is_server)
{
	SSLSrvConfigRec	*sc = mySrvConfig(s);
	SSL_CTX 		*ctx;
	char			*certfile;
#ifndef NO_DH
	static DH		*dh = NULL;
	BIO 			*bio = NULL;
#endif /* NO_DH */

	SSLeay_add_ssl_algorithms();
	SSL_load_error_strings();

	ctx = (is_server)?
			SSL_CTX_new(SSLv23_server_method()):
			SSL_CTX_new(SSLv23_client_method());

	if (is_server)  {
		if ((certfile = sc->szPublicCertFile[0]) == NULL)  {
			certfile = "";
			ap_log_error(APLOG_MARK, APLOG_WARNING, s,
						"No certificate file specified");
			}

#ifndef NO_RSA
		/* log(LOG_DEBUG, "Generating %d bit temporary RSA key...", KEYLENGTH);
		*/
#if SSLEAY_VERSION_NUMBER <= 0x0800
		rsa_tmp = RSA_generate_key(KEYLENGTH, RSA_F4, NULL);
#else
		rsa_tmp = RSA_generate_key(KEYLENGTH, RSA_F4, NULL, NULL);
#endif
		if (!rsa_tmp)  {
			ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_EMERG, s, "tmp_rsa_cb");
			exit (1);
			}
		/*  log(LOG_DEBUG, "Temporary RSA key generated");  */
		SSL_CTX_set_tmp_rsa_callback(ctx, tmp_rsa_cb);
#endif
#ifndef NO_DH
		if(!(bio = BIO_new_file(certfile, "r")))  {
			ap_log_error(APLOG_MARK, APLOG_WARNING, s,
						"DH: Could not read %s: %s", certfile,strerror(errno));
			goto dh_failed;
			}
		if (!(dh = PEM_read_bio_DHparams(bio, NULL, NULL
#if SSLEAY_VERSION_NUMBER >= 0x00904000L
			, NULL
#endif
			)))  {
			ap_log_error(APLOG_MARK, APLOG_WARNING, s,
						"Could not load DH parameters from %s", certfile);
			goto dh_failed;
			}
		SSL_CTX_set_tmp_dh(ctx, dh);
		/*	log (LOG_DEBUG, "Diffie-Hellman initialized with %d bit key",
		**			8*DH_size(dh));
		*/
		goto dh_done;
dh_failed:
		ap_log_error(APLOG_MARK, APLOG_WARNING, s,
					"Diffie-Hellman initialization failed");
dh_done:
		if (bio)	BIO_free(bio);
		if (dh) 	DH_free(dh);
#endif /* NO_DH */
		}

	SSL_CTX_set_session_cache_mode(ctx, SSL_SESS_CACHE_BOTH);

	/*	I think this is handled by ap directive SSLSessionCacheTimeout
	**	SSL_CTX_set_timeout(ctx, options.session_timeout);
	*/

    if (is_server)  {		/*  was options.option&OPT_CERT 	*/
        if (!SSL_CTX_use_certificate_file(ctx, certfile, SSL_FILETYPE_PEM))  {
            ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, s,
						"Error reading certificate file: %s", certfile);
            exit(1);
			}
        /*  log(LOG_DEBUG, "Certificate: %s", certfile);  */
#ifdef NO_RSA
        if (!SSL_CTX_use_PrivateKey_file(ctx, certfile, SSL_FILETYPE_PEM))  {
			ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, s,
						"SSL_CTX_use_PrivateKey_file");
            exit(1);
			}
#else /* NO_RSA */
        if (!SSL_CTX_use_RSAPrivateKey_file(ctx, certfile, SSL_FILETYPE_PEM)) {
            ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, s,
						"SSL_CTX_use_RSAPrivateKey_file");
            exit(1);
			}
#endif /* NO_RSA */
		}

	/*	options.verify_level & options.verify_use_only_my
	**	presumed to be handled by apache directives
	**	SSLVerifyClient & SSLVerifyDepth.
	*/

	if (sc->szCipherSuite  &&  strlen(sc->szCipherSuite)  &&
		!SSL_CTX_set_cipher_list(ctx, sc->szCipherSuite))  {
		ap_log_error(APLOG_MARK, APLOG_NOERRNO | APLOG_EMERG, s,
					"SSL_CTX_set_cipher_list");
		exit(1);
		}

	return ctx;
}


/*	Reinitializes channel and sets channel sock.
**	Closes ssl connection if it exists.
*/
void	reset_ch(Channel *chan, int fd)
{
	if (!chan)  return;

	if (chan->ssl)  {
		SSL_set_shutdown(chan->ssl, SSL_SENT_SHUTDOWN | SSL_RECEIVED_SHUTDOWN);
		SSL_free(chan->ssl);
		}
	
	chan->ssl = NULL;
	chan->ctx = NULL;
	chan->ptr = chan->bytes = 0;
	chan->fd = fd;
	chan->open = 1;

	return;
}


/*	Create SSL connection on (local or remote) channel to given host.
**	Return 0:Success.
*/
int 	new_ssl_ch(server_rec *s, Channel *chan, char *host)
{
	EVP_MD_CTX		hash_ctx;
	unsigned char	*sess_id;
	unsigned int 	len = getpid();

	if ((!chan->ctx  &&  !(chan->ctx = context_init(s, (host == NULL))))  ||
		(!chan->ssl  &&  !(chan->ssl = SSL_new(chan->ctx))))  {
        return 1;
		}

#if SSLEAY_VERSION_NUMBER >= 0x0922
	EVP_DigestInit(&hash_ctx, EVP_get_digestbyname("ssl3-sha1"));
	EVP_DigestUpdate(&hash_ctx, chan->ssl, sizeof(chan));
	EVP_DigestUpdate(&hash_ctx, chan, sizeof(chan));
	EVP_DigestUpdate(&hash_ctx, &len, sizeof(len));
	if (!(sess_id = (unsigned char *) malloc(EVP_MD_size(hash_ctx.digest))))  {
		return 2;
		}
	EVP_DigestFinal(&hash_ctx, sess_id, &len);
    SSL_set_session_id_context(chan->ssl, sess_id, len);
	free(sess_id);
#endif

#ifndef NO_KRB5
	if (chan->ssl->kssl_ctx)  {
		kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_SERVICE, KRB5SVC);
		kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_KEYTAB, KRB5KEYTAB);
		if (host)  kssl_ctx_setstring(chan->ssl->kssl_ctx, KSSL_SERVER, host);
		}
#endif	/* NO_KRB5 */

	return 0;
}


/*	Close any SSL channels.
*/
void	shutdown_ssl_ch(Channel *local, Channel *remote)
{
	/* 	have_cookie = 0;	*/

	if (local  &&  local->ssl)  {
		SSL_set_shutdown(local->ssl,  SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
		SSL_free(local->ssl);
		local->ssl = NULL;
		local->ctx = NULL;
		}
	if (remote  &&  remote->ssl)  {
		SSL_set_shutdown(remote->ssl, SSL_SENT_SHUTDOWN|SSL_RECEIVED_SHUTDOWN);
		SSL_free(remote->ssl);
		remote->ssl = NULL;
		remote->ctx = NULL;
		}
    ERR_remove_state(0);
}


/*	Do SSL read on given channel, return number of bytes read through *numptr.
**	Return 0:Success
*/
int 	rd_ssl_ch(Channel *chan, int *numptr)
{
	*numptr = SSL_read(chan->ssl, chan->buff + chan->ptr, BUFFSIZE - chan->ptr);

	switch (SSL_get_error(chan->ssl, *numptr))  {
		case SSL_ERROR_NONE:
			chan->ptr += *numptr;
			break;
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
			/*  log(LOG_DEBUG, "SSL_read returned WANT_ - retry");  */
			break;
		case SSL_ERROR_SYSCALL:
			if (*numptr)  { 	/* not EOF */
				/*	sockerror("SSL_read (socket)");	*/
				return 1;
			}
		case SSL_ERROR_ZERO_RETURN:
			/*  log(LOG_DEBUG, "SSL closed on read");  */
			chan->open = 0;
			break;
		case SSL_ERROR_SSL:
		default:
			/*	sslerror("SSL_read");	*/
			return 2;
		}

	return 0;
}


/*	Do SSL write on given channel, return #bytes written through *numptr.
**	Return 0:Success
*/
int 	wr_ssl_ch(Channel *from, Channel *to, int *numptr)
{
	*numptr = SSL_write(to->ssl, from->buff, from->ptr);

	switch (SSL_get_error(to->ssl, *numptr)) {
		case SSL_ERROR_NONE:
			memcpy(from->buff, from->buff + *numptr, from->ptr - *numptr);
			from->ptr -= *numptr;
			to->bytes += *numptr;
			break;
		case SSL_ERROR_WANT_WRITE:
		case SSL_ERROR_WANT_READ:
		case SSL_ERROR_WANT_X509_LOOKUP:
			/*  log(LOG_DEBUG, "SSL_write returned WANT_ - retry");  */
			break;
		case SSL_ERROR_SYSCALL:
			if (*numptr)  { /* not EOF */
				/*  sockerror("SSL_write (socket)");	*/
				return 1;
				}
		case SSL_ERROR_ZERO_RETURN:
			/*  log(LOG_DEBUG, "SSL closed on write");  */
			to->open = 0;
			break;
		case SSL_ERROR_SSL:
		default:
			/*  sslerror("SSL_write");  */
			return 2;
		}

	return 0;
}

